@@ -40,7 +40,7 @@ void distance_method_dialog(SEA_VIEW *view);
 matrix *run_calc_dist_matrix(allseq *seqs, int distkind, int protein, char **lwlseqs, int in_bootstrap, char **p_err_mess, int no_gui);
 char *run_distance_method(SEA_VIEW *view, int distkind, int remove_all_gaps, int use_bionj, int use_bootstrap,
 			 int replicates, Fl_Box *box, const char *distance_name, int using_kaks,
-			 const char* distancefname, const char *user_tree, int no_gui, void (*alert)(const char*, ...) );
+			 const char* distancefname, const char *user_tree, int no_gui, void (*alert)(const char*, ...), int keep_b_trees );
 char *bootstrap_reformat(char *tree, int replicates);
 matrix *Obs_Dist(allseq *data, model *mod);
 matrix *Kimura_p_Dist(allseq *data);
@@ -51,15 +51,15 @@ double hky(char* seq1, char* seq2, int lg, int *wght);
 int freq_obs(char* seq1, char* seq2, double* freq, int lgseq, int *wght);
 int informative(char** sequ, int nb);
 void jumble_fct(char** seq, char** name, int notu);
-char *run_parsimony_method(SEA_VIEW *view, int nogaps, const char *gapsunknown, int nreplicates, int njumbles, 
+static char *run_parsimony_method(SEA_VIEW *view, int nogaps, const char *gapsunknown, int nreplicates, int njumbles,
 			   int maxbest, int user_tree, Fl_Box *w_count, dnapars_S_option s_option,
-			   const char **pprogname, float consensus_threshold, int no_gui, void (*alert)(const char *,...) );
+			   const char **pprogname, float consensus_threshold, int no_gui, char **p_bootstrap_trees, void (*alert)(const char *,...) );
 void parsimony_dialog(SEA_VIEW *view);
 char *parsimony_with_jumble(char **pars_seqs, char **pars_names, int n_otu, int njumbles, int protein, int *psteps,
 	int *pcount, int maxbest, Fl_Widget *w_count, dnapars_S_option s_option, int no_gui, float consensus_threshold);
-char *parsimony_with_bootstrap(allseq *tmpallseq, char **pars_seqs, char **pars_names, int n_otu, int njumbles, 
+static char *parsimony_with_bootstrap(allseq *tmpallseq, char **pars_seqs, char **pars_names, int n_otu, int njumbles,
 	int protein, int *psteps, int maxbest, int replicates, Fl_Box *w_count, int *preplicates_done, dnapars_S_option s_option, int no_gui,
-			       float consensus_threshold);
+			       float consensus_threshold, int *pcount_best, char **p_bootstrap_trees);
 double *vector(long nl, long nh);
 void free_vector(double *v, long nl, long nh);
 int ludcmp(double **a, int n, int *indx, double *d);
@@ -69,13 +69,13 @@ matrix *LOGDET_dist(allseq *data);
 char *put_names_back_in_tree(char *oldtree, char **names);
 char *replace_tree_names_by_rank(const char *tree, char **names, int notu);
 void command_line_phylogeny(int argc, char **argv);
+void save_bootstrap_replicates(const char *fname, int replicates, SEA_VIEW *view);
 
 
 extern void interrupt_callback(Fl_Widget *ob, void *data);
 extern int save_phylip_file(const char *fname, char **seq,
 					 char **seqname, int totseqs, int *eachlength, region *region_used, 
 					 int *sel_seqs, int tot_sel_seqs, int phylipwidnames);
-extern Fl_Window *treedraw(char *tree, SEA_VIEW *view, const char *name, int from_tree_menu);
 extern int run_external_prog_in_pseudoterm(char *cmd, const char *dialogfname, const char *label);
 extern int init_add_bootstrap(char *full_tree);
 extern char *get_res_value(const char *name, const char *def_value);
@@ -198,7 +198,7 @@ void trees_callback(Fl_Widget *ob, void *data)
 		while(w != NULL) {
 			const char *c = w->xclass();
 			if(c != NULL && strcmp(c, TREE_WINDOW) == 0 &&
-				strcmp( ((FD_nj_plot*)w->user_data())->current_tree, p) == 0) {
+				strcmp( ((FD_nj_plot*)w->user_data())->trees, p) == 0) {
 					w->show();
 					return;
 				}
@@ -220,6 +220,17 @@ static void distance_callback(Fl_Widget *wgt, void *data)
 	else ((Fl_Widget *)data)->activate();
 }
 
+static void dist_b_cb(Fl_Widget *wgt, void *data)
+{
+  Fl_Button *me = (Fl_Button*)wgt;
+  Fl_Button *target = (Fl_Button*)data;
+  if (me->value()) target->activate();
+  else {
+    target->deactivate();
+    target->value(0);
+  }
+}
+
 void distance_method_dialog(SEA_VIEW *view)
 {
 	static int first = TRUE;
@@ -239,9 +250,10 @@ void distance_method_dialog(SEA_VIEW *view)
 	static Fl_Button *interrupt;
 	static Fl_Return_Button *go;
 	static int started, maxchoice;
+  static Fl_Check_Button *keep_b_trees;
 	if(first) {
 		first = FALSE;
-		w = new Fl_Window(250, 220);
+		w = new Fl_Window(250, 240);
 		w->label("Distance analysis");
 		w->set_modal();
 		nj = new Fl_Round_Button(3, 3, 60, 20, "NJ");
@@ -263,7 +275,11 @@ void distance_method_dialog(SEA_VIEW *view)
 		char tmp[10];
 		sprintf(tmp, "%d", def_replicates);
 		replicates->value(tmp);
-		box = new Fl_Box(3, bootstrap->y() + bootstrap->h() + 5, 150, 20, "");
+    keep_b_trees = new Fl_Check_Button(3, bootstrap->y() + bootstrap->h() + 5, 180, 20, "Show bootstrap trees");
+    keep_b_trees->deactivate();
+    bootstrap->callback(dist_b_cb, keep_b_trees);
+    bootstrap->when(FL_WHEN_CHANGED);
+		box = new Fl_Box(3, keep_b_trees->y() + keep_b_trees->h() + 5, 150, 20, "");
 		usertree = new Fl_Round_Button(nj->x(), box->y() + box->h() + 15, 85, nj->h(), "User tree:");
 		usertree->type(FL_RADIO_BUTTON);
 		treechoice = new Fl_Choice(usertree->x() + usertree->w(), usertree->y(), 155, 20, "Optimize branch lengths");
@@ -298,6 +314,7 @@ void distance_method_dialog(SEA_VIEW *view)
 	}
 	distance->value( FL_min(distance_choice, maxchoice) );
 	distance->callback(distance_callback, ignore_all_gaps);
+  keep_b_trees->value(0);
 	treechoice->clear();
 	vlength_menu *vmenu = view->menu_trees;
 	int count = view->menu_trees->vlength();
@@ -312,13 +329,26 @@ void distance_method_dialog(SEA_VIEW *view)
 	box->label("");
 	go->show();
 	w->show();
-	while(!started && !tree_build_interrupted) {
-		Fl_Widget *o = Fl::readqueue();
-		if (!o) Fl::wait();
-		else if(o == usertree || o == nj || o == bionj || o == savetofile) {
-			if( usertree->value() ) treechoice->activate(); else treechoice->deactivate();
-			}
-		}
+  while(!started && !tree_build_interrupted) {
+    Fl_Widget *o = Fl::readqueue();
+    if (!o) Fl::wait();
+    else if(o == usertree || o == nj || o == bionj || o == savetofile) {
+      if( usertree->value() || savetofile->value() ) {
+        bootstrap->value(0);
+        bootstrap->deactivate();
+        bootstrap->do_callback();
+      }
+      else {
+        bootstrap->activate();
+      }
+      if( usertree->value()  ) {
+        treechoice->activate();
+      }
+      else {
+        treechoice->deactivate();
+      }
+    }
+  }
 	distance_choice = distance->value();
 	int using_kaks = ((!view->protein) && (distance_choice == Ka || distance_choice == Ks) );
 	if(!tree_build_interrupted) {
@@ -332,7 +362,7 @@ void distance_method_dialog(SEA_VIEW *view)
 	  char *tree = run_distance_method(view, distance_choice, ignore_all_gaps->value(), bionj->value(), 
 						  bootstrap->value(), def_replicates, box, distance->text(), using_kaks,
 			      savetofile->value()?"":NULL, usertree->value() ? view->trees[treechoice->value()] : NULL,
-			      false, fl_alert);
+			      false, fl_alert, keep_b_trees->value());
 	  const char *choice = bionj->value()? "BioNJ_tree" : "NJ_tree";
 	  const char *n = extract_filename(view->masename);
 	  char *title = NULL;
@@ -344,7 +374,13 @@ void distance_method_dialog(SEA_VIEW *view)
 	    sprintf(title, "%s-%s", n, choice);
 	    if (q) *q = '.';
 	  }
-	  if (tree) treedraw(tree, view, n || usertree->value() ? title:choice, FALSE);
+    Fl_Window *treew = NULL;
+	  if (tree) treew = treedraw(tree, view, n || usertree->value() ? title:choice, FALSE);
+#if !defined(__APPLE__)
+    if (keep_b_trees->value() && treew) {
+      treew->position(treew->x() + 40, treew->y());
+    }
+#endif
 	  delete[] title;
 	}
 	w->hide();
@@ -402,10 +438,10 @@ matrix *run_calc_dist_matrix(allseq *seqs, int distkind, int protein, char **lwl
 
 char *run_distance_method(SEA_VIEW *view, int distkind, int remove_all_gaps, int use_bionj, int use_bootstrap,
 	int replicates, Fl_Box *box, const char *distance_name, int using_kaks, const char* distancefname,
-	const char *user_tree, int no_gui, void (*alert)(const char*, ...) )
+	const char *user_tree, int no_gui, void (*alert)(const char*, ...), int keep_b_trees )
 {
-	char *display_tree, *err_mess, tree_label[100]="";
-	int total = 0, i, j;
+	char *display_tree, *bootstrap_trees = NULL, *err_mess, tree_label[100]="";
+	int total = 0, i, j, lbtrees;
 	char **lwlseqs = NULL;
   	if(using_kaks) remove_all_gaps = FALSE;//keep gaps not to alter reading frame
 	allseq *phyml_seqs = view_to_allseq(view, remove_all_gaps);
@@ -476,14 +512,15 @@ char *run_distance_method(SEA_VIEW *view, int distkind, int remove_all_gaps, int
 	Bionj(phyml_mat);
 	sprintf(tree_label, "%s %d sites %s", use_bionj ? "BioNJ" : "NJ",
 			phyml_seqs->clean_len, distance_name );
-	if(use_bootstrap) {
+	if (use_bootstrap) {
 		char *full_tree = Write_Tree(phyml_mat->tree);
 		Free_Tree(phyml_mat->tree);
 		Free_Mat(phyml_mat);
 		int error = init_add_bootstrap(full_tree);
 		free(full_tree);
 		if(!error) {
-			for(int r = 0; r < replicates; r++) {
+      char *p;
+			for (int r = 0; r < replicates; r++) {
 				if (tree_build_interrupted) break;
 				bootstrap_weights(phyml_seqs, using_kaks );
 				phyml_mat = run_calc_dist_matrix(phyml_seqs, distkind, view->protein, lwlseqs, TRUE, &err_mess, no_gui);
@@ -495,6 +532,23 @@ char *run_distance_method(SEA_VIEW *view, int distkind, int remove_all_gaps, int
 				phyml_mat->tree = Make_Tree_From_Scratch(phyml_seqs->n_otu, phyml_seqs);
 				Bionj(phyml_mat);
 				char *one_replicate_tree = Write_Tree(phyml_mat->tree);
+        if (keep_b_trees) {
+          if (r == 0) {
+            lbtrees = replicates * (strlen(one_replicate_tree) + 1);
+            bootstrap_trees = (char *)malloc(lbtrees + 1);
+            p = bootstrap_trees;
+            *p = 0;
+          }
+          int l = strlen(one_replicate_tree);
+          if ((p-bootstrap_trees) + l + 1 > lbtrees) {
+            lbtrees = (p-bootstrap_trees) + l + 1000;
+            bootstrap_trees = (char*)realloc(bootstrap_trees, lbtrees + 1);
+            p = bootstrap_trees + strlen(bootstrap_trees);
+          }
+          memcpy(p, one_replicate_tree, l); p += l;
+          *p++ = '\n';
+          *p = 0;
+        }
 				Free_Tree(phyml_mat->tree);
 				Free_Mat(phyml_mat);
 				total += process_one_replicate(one_replicate_tree, 1.);
@@ -532,6 +586,9 @@ char *run_distance_method(SEA_VIEW *view, int distkind, int remove_all_gaps, int
 	sprintf(tree, "[%s] %s", tree_label, display_tree);
 	free(display_tree);
 	if (box) box->window()->hide();//necessary under X11 so tree is in foreground
+  if (bootstrap_trees) {
+    treedraw(bootstrap_trees, view, "bootstrap trees", FALSE);
+  }
   return tree;
 }
 
@@ -676,9 +733,20 @@ Fl_Round_Button *g_no; Fl_Choice *g_cats; Fl_Round_Button *a_est; Fl_Float_Input
 Fl_Round_Button *nni_b, *spr_b, *nni_spr_b;
 Fl_Round_Button *u_bionj, *u_menutree, *u_random;
 Fl_Choice *u_choice;
-Fl_Check_Button *u_optimize, *u_quiet;
+Fl_Check_Button *u_optimize, *u_quiet, *b_keep_trees;
 Fl_Int_Input *u_random_count;
 char *phyml_path;
+
+static void support_cb(Fl_Widget *wid, void *data)
+{
+  Fl_Button *b = (Fl_Button*)data;
+  if (b->value()) b_keep_trees->activate();
+  else {
+    b_keep_trees->deactivate();
+    b_keep_trees->value(0);
+  }
+}
+
 void phyml_dialog(SEA_VIEW *view)
 {
 	int y = 3;
@@ -763,7 +831,7 @@ void phyml_dialog(SEA_VIEW *view)
 		}
 	y += mymodel->h() + 25;
 	//bootstrap
-	Fl_Group *b_group = new Fl_Group(0, y, w->w(), 55, "Branch Support");
+	Fl_Group *b_group = new Fl_Group(0, y, w->w(), 75, "Branch Support");
 	b_group->box(FL_ROUNDED_BOX);
 	b_group->align(FL_ALIGN_TOP|FL_ALIGN_CENTER);
 	y += 5;
@@ -778,6 +846,12 @@ void phyml_dialog(SEA_VIEW *view)
 	b_count = new Fl_Int_Input(b_yes->x() + b_yes->w() + 5, y, 50, 20, "replicates");
 	b_count->align(FL_ALIGN_RIGHT);
 	b_count->static_value("100");
+  y += b_count->h();
+  b_keep_trees = new Fl_Check_Button(b_yes->x(), y, 180, 20, "Show bootstrap trees");
+  b_keep_trees->deactivate();
+  b_alrt->callback(support_cb, b_yes);
+  b_no->callback(support_cb, b_yes);
+  b_yes->callback(support_cb, b_yes);
 	b_group->end();
 	y += b_count->h() + 30;
   // freq + Ts/Tv group
@@ -958,8 +1032,8 @@ void run_phyml_callback(Fl_Widget *ob, void *data)
   strcpy(input, base_fname);
   strcat(input, ".phy");
   region r, *pr = &r;
+  list_segments ls;
   if(view->tot_sel_seqs > 0 && view->active_region == NULL) {
-	  list_segments ls;
 	  ls.debut = 1; ls.fin = view->seq_length; ls.next = NULL;
 	  r.list = &ls;
 	  }
@@ -1014,6 +1088,7 @@ void run_phyml_callback(Fl_Widget *ob, void *data)
   if(b_alrt->value()) replicates = -4;
   else if (b_no->value()) replicates = 0;
   else sscanf(b_count->value(), "%d", &replicates);
+  if (!u_optimize->value()) replicates = 0;
   if(i_no->value()) strcpy(invar, "0.0");
   else if(i_estim->value()) strcpy(invar, "e");
   else strcpy(invar, i_val->value());
@@ -1115,6 +1190,55 @@ void run_phyml_callback(Fl_Widget *ob, void *data)
 	  *p = 0;
 	  fclose(in);
 	  display_tree = put_names_back_in_tree(display_tree, truenames);
+    //process bootstrap trees
+    char *bootstrap_trees = NULL;
+    if (replicates > 0 && b_keep_trees->value()) {
+      sprintf(input, "%s.phy_phyml_boot_trees.txt", base_fname);
+      in = fopen(input, "r");
+      if (in) {
+        FILE *out = NULL;
+        sprintf(input, "%s.phy_phyml_boot_trees_truename.txt", base_fname);
+        out = fopen(input, "w");
+        long l = 1000;
+        while (out) { // for all bootstrap trees
+          char *b_tree, *q, *tree2;
+          int c;
+          b_tree = (char*)malloc(l + 1);
+          q = b_tree;
+          do { // read one line = one tree
+            c = fgetc(in);
+            if (c == EOF) {
+              free(b_tree);
+              break;
+            }
+            if (q - b_tree >= l) {
+              l += 1000;
+              tree2 = (char *)realloc(b_tree, l + 1);
+              q = tree2 + (q - b_tree);
+              b_tree = tree2;
+            }
+            *q++ = c;
+          } while (c != '\n');
+          if (c == EOF) break;
+          *q = 0;
+          b_tree = put_names_back_in_tree(b_tree, truenames);
+          fputs(b_tree, out);
+          free(b_tree);
+        }
+        fclose(in);
+        if (out) {
+          fclose(out);
+          in = fopen(input, "r");
+          fseek(in, 0, SEEK_END);
+          l = ftell(in);
+          fseek(in, 0, SEEK_SET);
+          bootstrap_trees = (char *)malloc(l+1);
+          fread(bootstrap_trees, l, 1, in);
+          bootstrap_trees[l] = 0;
+          fclose(in);
+        }
+      }
+    }
 	  //clean tree internal labels
 	  if(replicates > 0) display_tree = bootstrap_reformat(display_tree, replicates);
 	  else if(replicates < 0) display_tree = alrt_reformat(display_tree);
@@ -1138,7 +1262,14 @@ void run_phyml_callback(Fl_Widget *ob, void *data)
 	      sprintf(title, "%s-%s", n, "PhyML_tree");
 	      if (q) *q = '.';
 	    }
-	    treedraw(tree, view, n ? title:"PhyML_tree", FALSE);
+      Fl_Window *btreew, *treew;
+      if (bootstrap_trees) {
+        btreew = treedraw(bootstrap_trees, view, "bootstrap trees", FALSE);
+      }
+	    treew = treedraw(tree, view, n ? title:"PhyML_tree", FALSE);
+#if !defined(__APPLE__)
+      if (bootstrap_trees) treew->position(btreew->x() + 30, btreew->y() + 40);
+#endif
 	    delete[] title;
 	  }
 	}
@@ -1783,13 +1914,37 @@ int informative(char** sequ, int nb)
 		}
 }*/
 
-
-char *run_parsimony_method(SEA_VIEW *view, int nogaps, const char *gapsunknown, int nreplicates, int njumbles, 
-			   int maxbest, int user_tree, Fl_Box *w_count, dnapars_S_option s_option,
-			   const char **pprogname, float consensus_threshold, int no_gui, void (*alert)(const char *,...) )
+char *add_equally_best(char *tree, char *all_best, int count)
 {
-  char *tree = 0, *final_tree = 0;
-  int i, steps, info, count, replicates_done;
+  char *p, *q, *r, *final_tree;
+  final_tree = (char*)malloc(strlen(tree) + strlen(all_best)+ count * 30);
+  strcpy(final_tree, tree);
+  q = final_tree + strlen(final_tree);
+  *q++ = '\n';
+  p = all_best;
+  if (count > 1) {
+    int rank = 0;
+    while (count-- > 0) {
+      r = strchr(p, '\n');
+      sprintf(q, "[equally best tree #%d]", ++rank);
+      q += strlen(q);
+      memcpy(q, p, r - p + 1);
+      q += r-p+1;
+      p = r+1;
+    }
+  }
+  *q = 0;
+  free(tree);
+  free(all_best);
+  return final_tree;
+}
+
+static char *run_parsimony_method(SEA_VIEW *view, int nogaps, const char *gapsunknown, int nreplicates, int njumbles,
+			   int maxbest, int user_tree, Fl_Box *w_count, dnapars_S_option s_option,
+			   const char **pprogname, float consensus_threshold, int no_gui, char **p_bootstrap_trees, void (*alert)(const char *,...) )
+{
+  char *tree = 0, *final_tree = 0, *all_best;
+  int i, steps, info, count, replicates_done, count_best;
   const char *p;
   allseq *tmpallseqs = view_to_allseq(view, nogaps);
   if(*gapsunknown) gaps_as_unknown_states(tmpallseqs, view->protein);
@@ -1810,11 +1965,14 @@ char *run_parsimony_method(SEA_VIEW *view, int nogaps, const char *gapsunknown,
   }
   if (nreplicates) {
     tree = parsimony_with_bootstrap(tmpallseqs, pars_seqs, pars_names, tmpallseqs->n_otu, njumbles, 
-				    view->protein, &steps, maxbest, nreplicates, w_count, &replicates_done, s_option, no_gui, consensus_threshold);
+				    view->protein, &steps, maxbest, nreplicates, w_count, &replicates_done, s_option, no_gui, consensus_threshold, &count_best,
+                                    p_bootstrap_trees);
     if (tree) {
-      final_tree = (char *)malloc(strlen(tree) + 150);
-      sprintf(final_tree, "[%s, bootstrap with %d replic., %d steps, %d sites (%d informative)%s]%s",
-	      *pprogname, replicates_done, steps, tmpallseqs->clean_len, info, gapsunknown, tree);
+      final_tree = (char *)malloc(strlen(tree) + 200);
+      char best[35] = "";
+      if (count_best > 1) sprintf(best, "consensus of %d best trees, ", count_best);
+      sprintf(final_tree, "[%s, bootstrap with %d replic., %d steps, %s%d sites (%d informative)%s]%s",
+	      *pprogname, replicates_done, steps, best, tmpallseqs->clean_len, info, gapsunknown, tree);
     }
   }
   else if(user_tree) {
@@ -1837,18 +1995,23 @@ char *run_parsimony_method(SEA_VIEW *view, int nogaps, const char *gapsunknown,
       tree = parsimony(pars_seqs, pars_names, tmpallseqs->n_otu, 0, &steps, input_tree, maxbest, NULL, view->protein, s_option, no_gui, NULL);
       free(input_tree);
       if(tree == NULL) {
-	alert("Error in user-tree");
+        alert("Error in user-tree");
       }
       else {
-	final_tree = (char *)malloc(strlen(tree) + 150);
-	sprintf(final_tree, "[%s, user-tree, %d steps, %d sites (%d informative)%s]%s",
-		*pprogname, steps, tmpallseqs->clean_len, info, gapsunknown, tree);
+        final_tree = (char *)malloc(strlen(tree) + 150);
+        sprintf(final_tree, "[%s, user-tree, %d steps, %d sites (%d informative)%s]%s",
+                *pprogname, steps, tmpallseqs->clean_len, info, gapsunknown, tree);
       }
     }
   }
   else {
     tree = parsimony_with_jumble(pars_seqs, pars_names, tmpallseqs->n_otu, njumbles, view->protein, 
 				 &steps, &count, maxbest, w_count, s_option, no_gui, consensus_threshold);
+    if (count > 1) { // memorize all equally best trees in all_best string
+      char *q = strchr(tree, '\n');
+      all_best = strdup(q + 1);
+      *q = 0;
+    }
     if (tree && count > 1 && !view->protein) { /* with DNA, fit branch lengths to consensus of most parsimonious trees */
       char *input_tree = tree;
       int steps2; /* don't use the steps of this consensus tree because it can be larger than parsimonious steps */
@@ -1862,6 +2025,7 @@ char *run_parsimony_method(SEA_VIEW *view, int nogaps, const char *gapsunknown,
       else *aux = 0;
       sprintf(final_tree, "[%s, %d best trees%s, %d steps, %d sites (%d informative)%s]%s",
 	      *pprogname, count, aux, steps, tmpallseqs->clean_len, info, gapsunknown, tree);
+      if (count > 1) final_tree = add_equally_best(final_tree, all_best, count);
     }
   }
 way_out:
@@ -1872,12 +2036,18 @@ way_out:
   return final_tree;
 }
 
+static void user_tree_cb(Fl_Widget *wid, void *data)
+{
+  Fl_Button *b = (Fl_Button*)data;
+  b->value(0);
+  b->do_callback();
+}
 
 void parsimony_dialog(SEA_VIEW *view)
 {
 	int njumbles, i, count, maxbest = 100, started = 0, use_bootstrap, 
 		nreplicates, user_tree, nogaps; 
-	char *final_tree;
+	char *final_tree, *bootstrap_trees = NULL;
 	const char *p, *progname;
   dnapars_S_option s_option = more_thorough;
   
@@ -1930,10 +2100,16 @@ void parsimony_dialog(SEA_VIEW *view)
 											  40, 20, "replicates");
 	w_btrepl->value("100");
 	w_btrepl->align(FL_ALIGN_RIGHT);
-	Fl_Box *w_count = new Fl_Box(0, w_btrepl->y() + w_btrepl->h() + 5, w->w(), 20, NULL);
+  Fl_Check_Button *keep_b_trees = new Fl_Check_Button(2, w_dobootstrap->y() + w_dobootstrap->h() + 5, 180, 20, "Show bootstrap trees");
+  keep_b_trees->deactivate();
+  w_dobootstrap->callback(dist_b_cb, keep_b_trees);
+  w_dobootstrap->when(FL_WHEN_CHANGED);
+
+	Fl_Box *w_count = new Fl_Box(0, keep_b_trees->y() + keep_b_trees->h() + 5, w->w(), 20, NULL);
 	
 	Fl_Round_Button *w_user = new Fl_Round_Button(2, w_count->y() + w_count->h() + 5, 85, 20, "User tree:");
 	w_user->type(FL_RADIO_BUTTON);
+  w_user->callback(user_tree_cb, w_dobootstrap);
 	Fl_Choice *w_choice = new Fl_Choice(w_user->x() + w_user->w() + 5, w_user->y(), 130, 20, "");
 	count = view->menu_trees->vlength();
 	for( i = 0; i < count; i++) w_choice->add( view->menu_trees->vitem(i)->label() );
@@ -1994,7 +2170,7 @@ void parsimony_dialog(SEA_VIEW *view)
 	final_tree = run_parsimony_method(view, nogaps, gapsunknown, 
 				    use_bootstrap && nreplicates > 1 ? nreplicates : 0, njumbles, 
 					  maxbest, user_tree, w_count, s_option,
-					  &progname, slider->value()/100, false, fl_alert);
+                                    &progname, slider->value()/100, false, keep_b_trees->value() ? &bootstrap_trees : NULL, fl_alert);
 
 	fl_reset_cursor(w);
 	delete w;
@@ -2008,7 +2184,20 @@ void parsimony_dialog(SEA_VIEW *view)
 	    sprintf(title, "%s-%s", n, progname);
 	    if (q) *q = '.';
 	  }
-	  treedraw(final_tree, view, n ? title:progname, FALSE);
+    int tree_count = 0;
+    p = final_tree;
+    while (TRUE) {
+      p = strchr(p+1, '\n');
+      if (!p) break;
+      tree_count++;
+    }
+    if (tree_count == 0) tree_count = 1;
+    Fl_Window *btreew = NULL;
+    if (bootstrap_trees) btreew = treedraw(bootstrap_trees, view, "bootstrap trees", FALSE);
+	  Fl_Window *treew = treedraw(final_tree, view, n ? title:progname, FALSE, tree_count);
+#if !defined(__APPLE__)
+      if (btreew) treew->position(btreew->x() + 30, btreew->y() + 40);
+#endif
 	  delete[] title;
 	}
 }
@@ -2042,12 +2231,16 @@ char *parsimony_with_jumble(char **pars_seqs, char **pars_names, int n_otu, int
 			memcpy(newtree, p, q - p); newtree[q - p] = 0;
 			p = q + 1;
 		}
-		free(tree);
 		int **rows;
 		int br_count = after_consensus_items(consensus_threshold, count, &rows);
-		tree = compute_consensus_tree(pars_names, rows, br_count);
+		char *consensus_tree = compute_consensus_tree(pars_names, rows, br_count);
 		for (i = 0; i < br_count; i++) free(rows[i]);
 		free(rows);
+    char *newtree = (char*)malloc(strlen(consensus_tree) + strlen(tree) + 2);
+    sprintf(newtree, "%s\n%s", consensus_tree, tree);
+    free(consensus_tree);
+    free(tree);
+    tree = newtree;
 	}
 	*psteps = steps;
 	*pcount = count;
@@ -2055,17 +2248,22 @@ char *parsimony_with_jumble(char **pars_seqs, char **pars_names, int n_otu, int
 }
 
 
-char *parsimony_with_bootstrap(allseq *tmpallseqs, char **pars_seqs, char **pars_names, int n_otu, int njumbles, int protein, int *psteps,
+static char *parsimony_with_bootstrap(allseq *tmpallseqs, char **pars_seqs, char **pars_names, int n_otu, int njumbles, int protein, int *psteps,
 			int maxbest, int replicates, Fl_Box *w_count, int *preplicates_done, dnapars_S_option s_option, int no_gui,
-			       float consensus_threshold)
+			       float consensus_threshold, int *pcount_best, char **p_bootstrap_trees)
 {
-	int i, j, steps, newsteps, count = 0, replicates_done = 0;
-	char *tree, *newtree, *p;
+	int i, j, steps, newsteps, count = 0, replicates_done = 0, repl_count, lbtrees = 0;
+	char *tree, *newtree, *p, *q, *all_best, *bootstrap_trees = NULL;
 	
 	if (w_count) w_count->label("Full sequence tree"); 
 	tree = parsimony_with_jumble(pars_seqs, pars_names, n_otu, njumbles, protein, 
 								 &steps, &count, maxbest, NULL, s_option, no_gui, consensus_threshold);
 	if (!tree) return NULL;
+  if (count > 1) { // memorize all equally best trees in all_best string
+    p = strchr(tree, '\n');
+    all_best = strdup(p + 1);
+    *p = 0;
+  }
 	if (count > 1 && !protein) { /* with DNA, fit branch lengths to strict consensus of most parsimonious trees */
 	  char *input_tree = tree;
 	  /* don't use the steps of this consensus tree because it can be larger than parsimonious steps */
@@ -2084,14 +2282,32 @@ char *parsimony_with_bootstrap(allseq *tmpallseqs, char **pars_seqs, char **pars
 		bootstrap_weights(tmpallseqs, FALSE);
 		tree = parsimony(pars_seqs, pars_names, n_otu, njumbles, &steps, NULL, maxbest, tmpallseqs->wght, protein, s_option, no_gui, NULL);
 		if (!tree) break;
-		p = tree - 1; count = 0;
-		while((p = strchr(p+1, '\n')) != NULL) count++;//count equally parsimonious trees
-		i = count;
+		p = tree - 1; repl_count = 0;
+		while((p = strchr(p+1, '\n')) != NULL) repl_count++;//count equally parsimonious trees
+		i = repl_count;
 		while (TRUE) {
 			p = strchr(tree, '\n');
 			newtree = (char *)malloc(p - tree + 1);
 			memcpy(newtree, tree, p - tree); newtree[p - tree] = 0;
-			process_one_replicate(newtree, 1./count);
+			process_one_replicate(newtree, 1./repl_count);
+      if (p_bootstrap_trees) { // memorize all bootstrap trees
+        if (lbtrees == 0) {
+          lbtrees = 10000;
+          bootstrap_trees = (char *)malloc(lbtrees + 1);
+          q = bootstrap_trees;
+          *q = 0;
+        }
+        int l = strrchr(newtree, ')') - newtree + 1; // length of (....)
+        if ((q-bootstrap_trees) + l + 8 > lbtrees) {
+          lbtrees = (q-bootstrap_trees) + l + 8 + 10000;
+          char *tmpc = (char*)realloc(bootstrap_trees, lbtrees + 1);
+          q = tmpc + (q - bootstrap_trees);
+          bootstrap_trees = tmpc;
+        }
+        sprintf(q, "[%.2f]", 1./repl_count); q += strlen(q); // indicate tree weight
+        memcpy(q, newtree, l); q += l; // (...) part of tree
+        memcpy(q, ";\n", 3); q += 2; // end of tree
+      }
 			free(newtree);
 			i--;
 			if(i == 0) break;
@@ -2104,6 +2320,9 @@ char *parsimony_with_bootstrap(allseq *tmpallseqs, char **pars_seqs, char **pars
 		}
 	*preplicates_done = replicates_done;
 	tree = finish_add_bootstrap(replicates_done);
+  if (count > 1) tree = add_equally_best(tree, all_best, count);
+  *pcount_best = count;
+  if (p_bootstrap_trees) *p_bootstrap_trees = bootstrap_trees;
 	return tree;
 }
 
@@ -2171,14 +2390,14 @@ char *parsimony(char** seq, char** seqname, int notu, int njumbles, int *steps,
     args->done = false;
     if ( ! fl_create_thread((void (*) (void *))parsimony_thread, args, 1024) ) {
       while (!args->done) {
-	Fl::wait();
-	if (jumble_no > pre_jumble_no && jumble_no < njumbles && w_count) {
-	  pre_jumble_no = jumble_no;
-	  char tmp[40];
-	  sprintf(tmp, "Jumble attempt: %d steps: %d", jumble_no + 1, *args->steps); 
-	  w_count->label(tmp); 
-	}
-	}
+        Fl::wait();
+        if (jumble_no > pre_jumble_no && jumble_no < njumbles && w_count) {
+          pre_jumble_no = jumble_no;
+          char tmp[40];
+          sprintf(tmp, "Jumble attempt: %d steps: %d", jumble_no + 1, *args->steps);
+          w_count->label(tmp);
+        }
+      }
     }
     p = args->result;
     free(args);
@@ -2473,7 +2692,7 @@ void command_line_phylogeny(int argc, char **argv)
     if (distkind == 0) distname = strdup("Observed");
     view->tot_trees = 0;
     tree = run_distance_method(view, distkind, isarg(argc, argv, "-nogaps"), !isarg(argc, argv, "-NJ"), replicates > 1, replicates, NULL,
-			       distname, distkind == Ka || distkind == Ks, argname(argc, argv, "-distance_matrix"), NULL, true, Fl::fatal);
+			       distname, distkind == Ka || distkind == Ks, argname(argc, argv, "-distance_matrix"), NULL, true, Fl::fatal, false);
     if (isarg(argc, argv, "-distance_matrix")) exit(0);
   }
   if (isarg(argc, argv, "-parsimony")) { // parsimony method
@@ -2486,7 +2705,8 @@ void command_line_phylogeny(int argc, char **argv)
     else s_option = more_thorough;
     tree = run_parsimony_method(view, isarg(argc, argv, "-nogaps"), 
 				isarg(argc, argv, "-gaps_as_unknown") ? ", gaps treated as ?" : "", 
-				replicates, jumbles, 10000, 0, NULL, s_option, &progname, 1, true, Fl::fatal);
+				replicates, jumbles, 10000, 0, NULL, s_option, &progname, 1, true, 0, Fl::fatal);
+    if (tree && (p = strchr(tree, ';')) != NULL) *(p+1) = 0; // tree string can contain several trees, keep first only
   }
   if (tree) {
     char *outfname = argname(argc, argv, "-o");
@@ -2496,3 +2716,49 @@ void command_line_phylogeny(int argc, char **argv)
     if (out != stdout) fclose(out);
   }
 }
+
+void save_bootstrap_replicates(const char *fname, int replicates, SEA_VIEW *view)
+{
+  allseq *seqs = view_to_allseq(view, FALSE);
+  int *each_length = new int[seqs->n_otu];
+  for (int i = 0; i < seqs->n_otu; i++) each_length[i] = seqs->clean_len;
+  char **sequences = new char*[seqs->n_otu];
+  for (int i = 0; i < seqs->n_otu; i++) {
+    sequences[i] = new char[seqs->clean_len];
+  }
+  FILE *out = fl_fopen(fname, "w");
+  char *tmp_fname = create_tmp_filename();
+  for (int repl = 1; repl <= replicates; repl++) {
+    bootstrap_weights(seqs, FALSE);
+    for (int i = 0; i < seqs->n_otu; i++) {
+      int l = 0;
+      for (int j = 0; j < seqs->clean_len; j++) {
+        for (int k = 0; k < seqs->wght[j]; k++) {
+          sequences[i][l++] = seqs->c_seq[i]->state[j];
+        }
+      }
+    }
+    save_alignment_or_region(tmp_fname, sequences, NULL,
+                             NULL, view->seqname, seqs->n_otu, each_length,
+                             NULL, NULL, PHYLIP_FORMAT,
+                             0, NULL,
+                             NULL, NULL, 0, view->protein,
+                             0, NULL,
+                             NULL, view->phylipwidnames,
+                             0, NULL,
+                             NULL, view->spaces_in_fasta_names);
+    // add to growing output file
+    FILE *in = fl_fopen(tmp_fname, "r");
+    char line[100];
+    while (fgets(line, sizeof(line), in)) {
+      fputs(line, out);
+    }
+    fclose(in);
+    fl_unlink(tmp_fname);
+  }
+  fclose(out);
+  for (int i = 0; i < seqs->n_otu; i++) delete[] sequences[i];
+  delete[] sequences;
+  delete[] each_length;
+  free_after_view_to_allseq(seqs);
+}
